/***** construct.c ************************************************
 * Description: Routines for constructing a keyword tree.
 * Reference: Gusfield, D. (1997). Algorithms on Strings, Trees,
 *    and sequences; Computer Science and Computational Biology.
 *    CUP, p. 52ff.
 * Author: Bernhard Haubold, bernhard.haubold@fh-weihenstephan.de
 * File created on Wed Sep  8 10:20:28 2004.
 *
 * This file is part of medMatch.
 *
 *   medMatch is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   medMatch is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with medMatch; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************/
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "stringUtil.h"
#include "medMatch.h"

int nodeCounter;
int maxNodes;
KNode *tree;

/*
 * Generate keyword tree from non-symbol patterns in Pattern format. 
 * Word boundaries are always observed. 
 * This option introducts all possible combinations of the letters specified in the
 * variables "leading" and "lagging".
 */
KNode *createNonSymbolTree(Pattern *patternTab[], Args args, int numChar){
  KNode *root;
  int i, numLead, numLag;
  int strLen = 1000;
  Pattern *pp;
  char *leading = LEADING;
  char *lagging = LAGGING;
  char *str;

  nodeCounter = 1;

  if((str=(char *)malloc(strLen*sizeof(char))) == NULL){
    printf("ERROR [construct]: memory allocation failed\n");
    assert(0);
  }

  numLead = strlen(leading);
  numLag = strlen(lagging);

  maxNodes = numLead*(numChar+2)*numLag;
  /* tree = initializeTree(maxNodes); */
  /* root = tree; */
  root = getNode();

  for(i=0; i<HASHSIZE; i++)
    for(pp = patternTab[i]; pp != NULL; pp = pp->next)
      if(!(pp->pa->isAgs || pp->pa->isPs)){
	if(!args.A){                       /* add aliases only if they are not treated case-sensitively */
	  addModifiedPattern(pp,root,str);
	}else if(!(pp->pa->isA)){
	   addModifiedPattern(pp,root,str);
	}
      }

/*   tree = root; */
/*   if((tree = (KNode*) realloc(tree, (nodeCounter+1)*sizeof(KNode))) == NULL){ */
/*     printf("ERROR [construct]: reallocation of memory failed\n"); */
/*     assert(0); */
/*   } */
  preprocessTree(root);
  free(str);
  return root;
}

/*
 * Generate keyword tree from symbol patterns in Pattern format. 
 * Word boundaries are always observed. 
 * This option introducts all possible combinations of the letters specified in the
 * variables "leading" and "lagging".
 */
KNode *createSymbolTree(Pattern *patternTab[], Args args, int numChar){
  KNode *root;
  int i, numLead, numLag;
  int strLen = 1000;
  Pattern *pp;
  char *leading = LEADING;
  char *lagging = LAGGING;
  char *str;

  nodeCounter = 1;

  if((str=(char *)malloc(strLen*sizeof(char))) == NULL){
    printf("ERROR [construct]: memory allocation failed\n");
    assert(0);
  }
  numLead = strlen(leading);
  numLag = strlen(lagging);

  maxNodes = numLead*(numChar+2)*numLag;
  root = getNode();

  for(i=0; i<HASHSIZE; i++)
    for(pp = patternTab[i]; pp != NULL; pp = pp->next)
      if(pp->pa->isAgs || pp->pa->isPs)
	addModifiedPattern(pp, root, str);
      else if(pp->pa->isA && args.A)    /* add aliases only if they are treated case-sensitively */
	addModifiedPattern(pp, root, str);

  preprocessTree(root);
  free(str);
  return root;
}

/*
 * Generate keyword tree from patterns in Pattern format. 
 * Word boundaries are always observed. 
 * This option introducts all possible combinations of the letters specified in the
 * variables "leading" and "lagging".
 */
KNode *createPatternTree(Pattern *patternTab[], Args args, int numChar){
  KNode *root;
  int i, numLead, numLag;
  int strLen = 1000;
  Pattern *pp;
  char *leading = LEADING;
  char *lagging = LAGGING;
  char *str;

  nodeCounter = 1;

  if((str=(char *)malloc(strLen*sizeof(char))) == NULL){
    printf("ERROR [construct]: memory allocation failed\n");
    assert(0);
  }

  numLead = strlen(leading);
  numLag = strlen(lagging);

  maxNodes = numLead*(numChar+2)*numLag;
  /* tree = initializeTree(maxNodes); */
  /* root = tree; */
  root = getNode();

  for(i=0; i<HASHSIZE; i++)
    for(pp = patternTab[i]; pp != NULL; pp = pp->next)
      addModifiedPattern(pp,root,str);

  preprocessTree(root);
  free(str);
  return root;
}

/* add borders to string before inserting it into the pattern table */
void addModifiedPattern(Pattern *pp, KNode *root, char *str){
  int l, j, k, numLead, numLag;
  char *leading = LEADING;
  char *lagging = LAGGING;

  numLead = strlen(leading);
  numLag = strlen(lagging);

  l = strlen(pp->str);
  for(j=0; j<l; j++)
    str[j+1]=pp->str[j];
  for(j=0;j<numLead;j++){
    str[0] = leading[j];
    for(k=0;k<numLag;k++){
      str[l+1]=lagging[k];
      str[l+2]='\0';
      addPattern(pp,str,root);
    }
  }
}

/*
 * Add a pattern to the keyword tree. Because we are dealing with 
 * a trie we cannot treat every node as branching.
 */
void addPattern(Pattern *pp, char *str, KNode *root){
  KNode *np, *child, *parent;
  int stringLength;
  int depth, count;
  depth = 0;
  np = root; 

  stringLength= strlen(str);

  /* printf("addString %s %d\n",str,strlen(str)); */

  parent = np;
  while((np = findChild(np,*str)) != NULL && depth < stringLength){
    str++;
    depth++;
    parent = np;
  }
  if(depth == stringLength && np != NULL){
    /* make sure pattern list is non-redundant */
    if(np->p == NULL){
      np->p = pp;
    }else{
      printf("ERROR [keywordTree]: Attempted to add two identical strings to keyword tree.\n");
      printf("                     The program only accepts non-redundant pattern lists.\n");
      exit(-1);
    }
  }else{
    np = parent;
    count = 0;
    while(depth < stringLength){
      if(nodeCounter>=maxNodes){
	printf("ERROR [keywordTree]: Ran out of nodes at node #%d; max. nodes: %d!\n",nodeCounter,maxNodes);
	assert(0);
      }
      /* child = ++tree; */
      child = getNode();
      nodeCounter++;
      child->c = *str++;
      child->depth = ++depth;
      addChild(parent,child);
      parent = child;
    }
    parent->p = pp;
  }
}

void addChild(KNode *parent, KNode *child){
  KNode *np;

  if(parent->child == NULL){
    parent->child = child;
  }else{
    np = parent->child;
    while(np->sibling != NULL){
      np = np->sibling;
    }
    np->sibling = child;
  }
}

/*
 * Purpose: Find a child node of parent whose edge is labelled with c.
 */
KNode *findChild(KNode *parent, char c){
  KNode *np;

  if(parent->child == NULL){
    return NULL;
  }
  np = parent->child;
  while(np->c != c){
    if(np->sibling == NULL){
      break;
    }
      np = np->sibling;
  }
  if(np->c == c){
    return np;
  }else{
    return NULL;
  }
}


/*
 * Purpose: Initialize keyword tree.
 * Parameter:
 *   int numChar: number of characters contained in pattern set.
 * Returns: pointer to allocated array of type node.
 */
KNode *initializeTree(int numChar){
  int i;
  /* allocate memory for tree to the maximal size possible */
  if((tree = (KNode *) malloc((numChar+1)*sizeof(KNode))) == NULL){
    printf("ERROR [keywordTree]: memory allocation failed.\n");
    assert(0);
  }
  /* make sure that all nodes are initialized to sensible values */
  for(i=0; i<=numChar; i++){ 
    tree[i].p = NULL;
    tree[i].depth = 0;
    tree[i].child = NULL;
    tree[i].sibling = NULL;
    tree[i].fail = NULL;
    tree[i].out = NULL;
  }
  tree[0].c='\0';
  return tree;
}

KNode *getNode(){
  KNode *node;
  node = (KNode *) malloc(sizeof(KNode));
  node->c = '\0';
  node->p = NULL;
  node->depth = 0;
  node->child = NULL;
  node->sibling = NULL;
  node->fail = NULL;
  node->out = NULL;
  return node;
}
