/***** constructPatternTable.c ************************************
 * Description: Construct hash table containing all patterns to be 
 *              searched for. Each pattern is associated with a
 *              binary tree containing in each node the reference
 *              to a gene, as well as the attributes of the word
 *              in that particular gene.
 * Author: Bernhard Haubold, bernhard.haubold@fh-weihenstephan.de
 * File created on Sat Oct 23 10:26:23 2004.
 *
 * This file is part of medMatch.
 *
 *   medMatch is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   medMatch is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with medMatch; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <math.h>
#include "eprintf.h"
#include "medMatch.h"
#include "stringUtil.h"

Pattern **computeScore(Pattern **hashTab);
static char *fields[MAXFIELDS];
/*
 * Returns a pointer to an array of patterns. Genes are only accessed via patterns.
 */
Pattern **constructPatternTable(FILE *fp, Args args, Pattern *hashTab[], int *numChar, Gene *genes){
  Pattern *pattern;
  static Pattern *patternTab[HASHSIZE];
  /* FILE *gfp; */
  int numFields, numWords;
  int i = 0, lower = 1;
  int geneCount = 0;
  int gi;
  char *line;
  static char *words[MAXFIELDS];
  Indices *indices;

  for(i=0; i<HASHSIZE; i++){
    patternTab[i] = NULL;
  }

  /* *fields = (char *) malloc(MAXFIELDS*sizeof(char)); */

  *numChar = 0;
  hashTab = patternTab;
  /* mark non-existent genes and set counters of various categories to zero */
  for(i=0; i<MAX_GENES; i++){
    genes[i].id = 0;
    genes[i].countAgn = 0;
    genes[i].countPgn = 0;
    genes[i].countAlias = 0;
    genes[i].lengthPs = 0;
    genes[i].lengthAgs = 0;
    genes[i].sym = NULL;
  }

  /* allocate space for line of text */
  if((line = (char *)malloc(MAXLINE*sizeof(char))) == NULL){
    printf("ERROR [medMatch]: memory allocation failed.\n");
    assert(0);
  }

  /* make sure that file is open */
  if(fp == NULL){
    printf("ERROR [medMatch]: passed void file pointer\n");
    assert(0);
  }
  /* get field indices from header line */
  if((line = fgets(line,MAXLINE,fp)) == NULL){
    printf("ERROR [medMatch]: empty input file?\n");
    assert(0);
  }
  /* get indices for required fields */
  indices = (Indices *)emalloc(sizeof(Indices));
  getIndices(line, indices);
  /* loop over remaining lines in input file; each line corresponds to one gene */
  i=0;
  while((line = fgets(line,MAXLINE,fp)) != NULL){
    /* split line along tabs */
    if(strpbrk(line,"\t") == NULL)
      continue;
    split(line, "\t", fields, &numFields);
    if(numFields < 2){
      printf("numFields: %d\n",numFields);
      printf("%s\n",line);
      printf("ERROR [medMatch]: the input file appears to contain lines with fields\n");
      printf("                  not separated by tabs. Please reformat your input file\n");
      printf("                  using, for example, the command\n");
      printf("                  perl -pe 's/\\s\\s+/\\t/g' inputFile > outputFile\n");
      exit(-1);
    }

    /* check that we are dealing with an approved gene */
    if(fields[indices->st] == NULL)
      continue;
    if((strstr(fields[indices->st],"Approved"))==NULL)
      continue;
    /* get id */
    if(!isdigit(fields[indices->hg][0]))
      continue;
    /* printf("cp - 5\n"); */
    gi = atoi(fields[indices->hg]);
    genes[gi].isNew = 1;
    genes[gi].id = gi;
    /* process approved gene name */
    lower = 1; /* change pattern to lower case */
    split(fields[indices->an]," ",words,&numWords);
    genes[gi].wordsAgn = 0;
    for(i=0;i<numWords;i++){
      words[i] = cleanWordEdges(words[i]);
      words[i] = cleanWord(words[i]);
      if(testWord(words[i],MIN_NUM_CHAR,MIN_WORD_LEN)){
	pattern = install(words[i],hashTab,lower);
	pattern->scoreAgn++;
	pattern->pa = addPatternAttribute(pattern,pattern->pa, genes+gi, ISAGN);
	*numChar += strlen(words[i]);
	words[genes[gi].wordsAgn++] = words[i];
      }
    } 
    genes[gi].wordsAgn = countDistinctWords(words,genes[gi].wordsAgn);
    /* install approved gene symbol */
    lower = 0;
    if(fields[indices->as] != NULL){
      genes[gi].sym = fields[indices->as];
      pattern = install(fields[indices->as],hashTab,lower);
      pattern->pa = addPatternAttribute(pattern,pattern->pa, genes+gi, ISAGS);
      *numChar += strlen(words[0]);
    }
    /* install previous symbols */
    lower = 0;
    if(fields[indices->ps] != NULL){
      split(fields[indices->ps],",", words, &numWords);
      genes[gi].lengthPs = 0;
      for(i=0;i<numWords;i++){
	if(strlen(words[i])>0){
	  words[i] = rmBlanks(words[i]);
	  pattern = install(words[i],hashTab,lower);
	  pattern->scorePgn++;
	  pattern->pa = addPatternAttribute(pattern,pattern->pa, genes+gi, ISPS);
	  *numChar += strlen(words[i]);
	}
      } 
    }
    /* install aliases */
    if(args.A)
      lower = 0;
    else
      lower = 1;
    if(fields[indices->al] != NULL){
      split(fields[indices->al],",", words, &numWords);
      for(i=0;i<numWords;i++){
	if(words[i] != NULL){
	  words[i] = cleanWordEdges(words[i]);
	  words[i] = cleanWord(words[i]);
	  if(testWord(words[i],MIN_NUM_CHAR,MIN_SYMBOL_LEN)){
	    pattern = install(words[i],hashTab,lower);
	    pattern->pa = addPatternAttribute(pattern,pattern->pa, genes+gi, ISA);
	    *numChar += strlen(words[i]);
	  }
	}
      } 
    }
    /* install previous gene name */
    lower = 1;
    if(fields[indices->pn] != NULL){
      split(fields[indices->pn]," ",words,&numWords);
      genes[gi].wordsPgn = 0;
      for(i=0;i<numWords;i++){
	if(words[i] != NULL){
	  words[i] = cleanWordEdges(words[i]);
	  words[i] = cleanWord(words[i]);
	  if(testWord(words[i],MIN_NUM_CHAR,MIN_WORD_LEN)){
	    pattern = install(words[i],hashTab,lower);
	    pattern->pa = addPatternAttribute(pattern,pattern->pa, genes+gi, ISPGN);
	    *numChar += strlen(words[i]);
	    words[genes[gi].wordsPgn++] = words[i];
	  }
	}
      }
      genes[gi].wordsPgn = countDistinctWords(words,genes[gi].wordsPgn);
    }
    geneCount++;
  }
  hashTab = computeScore(hashTab);
  free(line);
  return hashTab;
}

Pattern **computeScore(Pattern **hashTab){
  Pattern *p;
  /* float x; */
  int i;

  for(i=0;i<HASHSIZE;i++){
    p = hashTab[i];
    for(p=hashTab[i];p != NULL; p = p->next){
      /* p->scoreAgn = 1; */
      /* p->scorePgn = 1; */
      if(p->scoreAgn >= 5)
	p->scoreAgn = 0;
      else
	p->scoreAgn = 1;
      if(p->scorePgn >= 5)
	p->scorePgn = 0;
      else
	p->scorePgn = 1;

/*       x = 1.0/p->scoreAgn; */
/*       p->scoreAgn = (x > 0) ? x : 0; */
/*       x = 1.0/p->scorePgn; */
/*       p->scorePgn = (x > 0) ? x : 0; */
    }
  }
  return hashTab;
}

/* lookup: look for pattern in hashtab; patterns are indexed by their pattern string */
Pattern *lookup(char *word, Pattern *hashTab[]){
  Pattern *pp;

  for(pp = hashTab[hash(word)]; pp != NULL; pp = pp->next)
    if(strcmp(word,pp->str) == 0)
      return pp;     /* found */

  return NULL;       /* not found */
  
}

/* install: adds a new Pattern corresponding to a pattern string to the hashTab and
 *          returns a pointer to the new Pattern. If the pattern string is already contained
 *          in the hashtab, a pointer to to the corresponding Pattern is returned. */
Pattern *install(char *str, Pattern *hashTab[], int lower){
  Pattern *pp;
  unsigned hashval;

  if(str[0] == '\0'){
    printf("INSTALLING NULL!!!!!!!!!!!!\n");
    exit(0);
  }
  if(lower)
    str = strtolower(str);

  if((pp = lookup(str,hashTab)) == NULL){ /* not found */
    pp = (Pattern *) malloc(sizeof(*pp));
    pp->pa = NULL;
    if(pp == NULL || (pp->str = strdup2(str)) == NULL)
      return NULL;
    hashval = hash(str);
    pp->next = hashTab[hashval];
    hashTab[hashval] = pp;
    /* set pointer to pattern attributes to null */
    pp->pa = NULL;
    pp->scoreAgn = 0;
    pp->scorePgn = 0;
  }
  return pp;                             /* already there */
}

/* return 1 if
 * a) word contains >= 1 numLet and
 * b) contains >= length characters
 */
int testWord(char *word, int numLet, int length){
  int lc, cc;
  char c;
  
  /* set length counter and character counter to zero */
  lc = cc = 0;
  
  while((c = *word) != '\0'){
    word++;
    lc++;
    if(isalpha(c))
      cc++;
  }
  if(cc>=numLet && lc>=length)
    return 1;
  else
    return 0;
}


/* count the number of distinct words in  a list of words of length num */
int countDistinctWords(char *words[], int num){
  int i, j, k, c, found;
  found = 0;
  c = num;
  k = num - 1;

  for(i=0;i<k;i++){
    for(j=i+1;j<num;j++){
      if(strcmp(words[i],words[j]) == 0){
	found = 1;
	break;
      }
    }
    if(found){
      c--;
      found = 0;
    }
  }
  return c;
}


/* getIndices: get indices of fields and check that all fields
 * necessary for this analysis are filled
 */
void getIndices(char *line, Indices *indices){
  int numFields;
  int i, error;
  /* initialize indices */
  indices->hg = -1;
  indices->an = -1;
  indices->as = -1;
  indices->ps = -1;
  indices->al = -1;
  indices->pn = -1;
  indices->st = -1;
  /* find required fields */
  split(strtolower(line), "\t", fields, &numFields);  
  for(i=0;i<numFields;i++){
    if(strstr(fields[i],"hgnc id")){
      indices->hg = i;
    }else if(strstr(fields[i],"approved name"))
      indices->an = i;
    else if(strstr(fields[i],"approved symbol"))
      indices->as = i;
    else if(strstr(fields[i],"previous symbols"))
      indices->ps = i;
    else if(strstr(fields[i],"aliases"))
      indices->al = i;
    else if(strstr(fields[i],"previous names"))
      indices->pn = i;
    else if(strstr(fields[i],"status"))
      indices->st = i;
  }
  /* check for missing fields */
  error = 0;
  if(indices->hg == -1){
    printf("ERROR: column \"HGNC ID\" is missing from pattern file.\n");
    error = 1;
  }
  if(indices->an == -1){
    printf("ERROR: column \"Approved Name\" is missing from pattern file.\n");
    error = 1;
  }
  if(indices->as == -1){
    printf("ERROR: column \"Approved Symbol\" is missing from pattern file.\n");
    error = 1;
  }
  if(indices->ps == -1){
    printf("ERROR: column \"Previous Symbols\" is missing from pattern file.\n");
    error = 1;
  }
  if(indices->al == -1){
    printf("ERROR: column \"Aliases\" is missing from pattern file.\n");
    error = 1;
  }
  if(indices->pn == -1){
    printf("ERROR: column \"Previous Names\" is missing from pattern file.\n");
    error = 1;
  }
  if(indices->st == -1){
    printf("ERROR: column \"Status\" is missing from pattern file.\n");
    error = 1;
  }
  if(error)
    exit(-1);
}
