/***** medMatch.h ************************************************
 * Description: Header file for medMatch program.
 * Author: Bernhard Haubold, bernhard.haubold@fh-weihenstephan.de
 * File created on Wed Sep  8 10:20:51 2004.
 *****************************************************************/
#include "sequenceData.h"
#include "interface.h"

#define ISAGN 0  /* approved gene name */
#define ISAGS 1  /* approved gene symbol */
#define ISPS  2  /* previous symbol */
#define ISA   3  /* alias */
#define ISPGN 4  /* previous gene name */

#define MIN_WORD_LEN 4              /* minimum length of word for inclusion in pattern list */
#define MIN_SYMBOL_LEN 3            /* minimum length of symbol for inclusion in pattern list */
#define MIN_NUM_CHAR 1              /* minimum number of characters for inclusion in pattern list */
#define MAXNUMWORD 20000            /* maximum number of words that can be detected per article */
#define MAX_WORD_COUNT 10           /* maximum count per article permitted for a word */
#define MAXWORDGENES 20             /* maximum number of genes based on word matches */
#define MAX_PATTERNS 40000          /* maximum number of patterns */
#define MAXLINE 1000000             /* maximum length of a line that can be read in */
#define MAX_GENES 50000             /* maximum number of genes that can be processed */
#define LEADING ">([ \"\'"          /* list of leading characters */
#define LAGGING "<)] .,:;?!\"\'/"   /* list of lagging characters */


typedef struct gene{ /* representation of a gene */
  int id;            /* HGNC */
  char *sym;         /* gene symbol */
  int wordsAgn;      /* number of words in approved gene name */
  int wordsPgn;      /* number of words in previous gene name */
  int countAgn;
  int countPgn;
  int countAlias;
  int lengthAgs;
  int lengthPs;      /* maximal length of the previous symbols seen */
  unsigned int isNew : 1;
  short matchCategory;
} Gene;

typedef struct pattern{          /* a pattern, i.e. a string that is searched for in medLine */
  char *str;                     /* the string */
  struct patternAttributes *pa;  /* pointer to the binary tree of pattern attributes */
  struct pattern *next;          /* define next entry in chain */
  float scoreAgn;                /* score of this pattern in the category "approved gene name */
  float scorePgn;                /* score of this pattern in the category "previous gene name */
} Pattern;

typedef struct patternAttributes{  /* the attributes of a pattern w.r.t a gene; forms a binary search tree
				    * ordered w.r.t the id of the gene */
  struct gene *gene;               /* pointer to the gene concerned */
  struct pattern *pattern;         /* pointer to pattern of origin */
  struct patternAttributes *left;  /* pointer to the left child */
  struct patternAttributes *right; /* pointer to the right child */
  unsigned int isAgn : 1;          /* is part of an approved gene name */
  unsigned int isAgs : 1;          /* is an approved gene symbol */
  unsigned int isPs  : 1;          /* is a previous symbol */
  unsigned int isA   : 1;          /* is an alias */
  unsigned int isPgn : 1;          /* is part of a previous gene name */
} PatternAttributes;


typedef struct node{   /* node in a keyword tree */
  Pattern *p;          /* pointer to pattern */
  int depth;           /* number of characters on path from root to node */
  char c;              /* character labelling incoming edge */
  struct node *fail;   /* failure link */
  struct node *out;    /* output link */
  struct node *child;
  struct node *sibling;
} KNode; 


typedef struct wordNode{    /* node in a word tree */
    Pattern *pattern;       /* points to the pattern */
    int count;              /* number of occurrences */
    struct wordNode *left;  /* left child */
    struct wordNode *right; /* right child */
} WordNode;

typedef struct geneNode{    /* node in a gene tree */
  Gene *gene;
  int agn;
  int pgn;
  int symbolLength;
  int alias;
  int ps;
  struct geneNode *left;
  struct geneNode *right;
} GeneNode;

typedef struct indices{ /* indexes of fields in synonymes file */
  int hg;   /* hgnc id index */
  int an;   /* approved names index */
  int as;   /* approved symbols index */
  int ps;   /* previous symbos index */
  int al;   /* aliases index */
  int pn;   /* previous names index */
  int st;   /* status index */
} Indices;

KNode *initializeTree(int numChar);
KNode *createSequenceKeywordTree(Sequence *patterns, Args args, int numPat, int numChar);
KNode *createKeywordTree(char **patterns, int numPat, int numChar);
void preprocessTree(KNode *root);
void createFailureLink(KNode *source, KNode *target);
KNode *compareAndAdjust(KNode *source, KNode *target, int *found);
void addString(char *str, char *pattern, KNode *root);
void addPattern(Pattern *pp, char *str, KNode *root);
void addModifiedPattern(Pattern *pp, KNode *root, char *str);
void addChild(KNode *parent, KNode *child);
KNode *findChild(KNode *parent, char c);
void printTree(KNode *root);
void examineFailureLinks(KNode *source);
void printTreeBreadthFirst(KNode *root);
void search(Sequence *patterns, char *text, KNode *root);
void searchOnline(Sequence *patterns, char *fileName, KNode *root, Args args);
void searchMedline(Pattern *patternTab[], char *fileName, KNode *symbolRoot, KNode *nonSymbolRoot, Args args, Gene *genes);
WordNode *addWord(WordNode *p, Pattern *pattern, WordNode *wordNodePool);
void harvestTree(WordNode *p);
WordNode *searchLine(char *text, KNode *root, WordNode *wordRoot, WordNode *wordNodePool);
char **processPatterns(char *patterns, int *patternSize, int *residues);
char **readPatterns(char *file, char *patterns[], int *patternSize, int *residues);
Pattern *install(char *str, Pattern *hashTab[], int lower);
Pattern *lookup(char *word, Pattern *hashTab[]);
int testWord(char *word, int numLet, int length);
Pattern **constructPatternTable(FILE *fp, Args args, Pattern *hashTab[], int *numChar, Gene *genes);
KNode *createPatternTree( Pattern *patternTab[], Args args, int numChar);
KNode *getNode();
void getIndices(char *line, Indices *indices);

void addGene(PatternAttributes *pa);
void harvestPatternAttributes(PatternAttributes *pa);
int getApprovedCount(Args args);
KNode *createNonSymbolTree(Pattern *patternTab[], Args args, int numChar);
KNode *createSymbolTree(Pattern *patternTab[], Args args, int numChar);
void resetGenes(int count, int *ind);
int countDistinctWords(char *words[], int num);

/* functions in attributeTree.c */
PatternAttributes *addPatternAttribute(Pattern *pattern, PatternAttributes *pa, Gene *gene, int attr);
PatternAttributes *paAlloc(void);
PatternAttributes *setAttribute(PatternAttributes *pa, int attr);
void testTree(PatternAttributes *pa);
